/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Yacc extension's configuration parser.
 * See doc/rt_default.conf.sample for configuration file format
 */

/* For development only : */
%debug 
%error-verbose

/* The parser receives the configuration file filename as parameter */
%parse-param {char * conffile}

/* Keep track of location */
%locations 
%pure-parser

%{
#include "rt_default.h"
#include "rtd_gram.h"	/* bison is not smart enough to define the YYLTYPE before including this code, so... */

#include <string.h>
#include <errno.h>

/* Forward declaration */
int yyparse(char * conffile);

/* Parse the configuration file */
int conf_parse(char * conffile)
{
	extern FILE * yyin;
	int ret;
	
	TRACE_ENTRY("%p", conffile);
	
	TRACE_DEBUG (FULL, "Parsing configuration file: %s...", conffile);
	
	yyin = fopen(conffile, "r");
	if (yyin == NULL) {
		ret = errno;
		log_error("Unable to open extension configuration file %s for reading: %s\n", conffile, strerror(ret));
		TRACE_DEBUG (INFO, "Error occurred, message logged -- configuration file.");
		return ret;
	}

	ret = yyparse(conffile);

	fclose(yyin);

	if (ret != 0) {
		TRACE_DEBUG (INFO, "Unable to parse the configuration file.");
		return EINVAL;
	}
	
	return 0;
}

/* The Lex parser prototype */
int yylex(YYSTYPE *lvalp, YYLTYPE *llocp);

/* This function checks a string value can be a DiameterId (== a fqdn) */
static int is_valid_fqdn( char * candidate ) 
{
	/* We first search for a '.' */
	if (!strchr(candidate, '.')) {
		log_error("The string '%s' is not a valid fully-qualified domain name (fqdn).\n", candidate);
		return 0;
	}
	
	/* We may do additional checking here */
	
	/* Ok this candidate is valid */
	return 1;
}

/* This function checks a string value can be a Realm */
static int is_valid_realm( char * candidate ) 
{
	/* What are the constraints? i18n and so on... See for this another time... */
	
	/* Ok this candidate is valid */
	return 1;
}

/* Function to report the errors */
void yyerror (YYLTYPE *ploc, char * conffile, char const *s)
{
	TRACE_DEBUG(INFO, "Error in configuration parsing");
	
	if (ploc->first_line != ploc->last_line)
		log_error("%s:%d.%d-%d.%d : %s\n", conffile, ploc->first_line, ploc->first_column, ploc->last_line, ploc->last_column, s);
	else if (ploc->first_column != ploc->last_column)
		log_error("%s:%d.%d-%d : %s\n", conffile, ploc->first_line, ploc->first_column, ploc->last_column, s);
	else
		log_error("%s:%d.%d : %s\n", conffile, ploc->first_line, ploc->first_column, s);
}

/* Global var to simplify reporting prefered route */
int pref;

%}

/* Values returned by lex for token */
%union {
	char 		*string;	/* The string is allocated by strdup in lex.*/
	int		 integer;	/* Store integer values */
}

/* In case of error in the lexical analysis */
%token 		LEX_ERROR

/* The debug level token */
%token 		VERBOSITY

/* Tokens to overwrite default configuration */
%token 		HANDLE_DESTINATION_HOST
%token 		HANDLE_DESTINATION_REALM
%token 		SCORE_DESTINATION_HOST
%token 		SCORE_DESTINATION_REALM
%token 		SCORE_NO_INFO
%token 		SCORE_DEFAULT_PEER
%token 		SCORE_DEFAULT_PEER_PREF
%token 		SCORE_REALM
%token 		SCORE_REALM_PREF

/* Tokens and types for routing table definition */
/* A (de)quoted string (malloc'd in lex parser; it must be freed after use) */
%token <string>	QSTRING
/* used in these subtypes */
%type <string>	FQDN		/* This is a fqdn. We check that the syntax is correct. */
%type <string>	REALM		/* This is a realm. */
%type <string>	ROUTE		/* A realm or NULL for default realm */
%type <string>	DEST		/* the FQDN, and the "pref" global var is defined */

/* An integer value */
%token <integer> INTEGER



/* -------------------------------------- */
%%

	/* The grammar definition */
conffile:		/* empty grammar is OK */
			| conffile verbosity
			| conffile confhdl
			| conffile confscore
			| conffile peerdef
			;

	/* Defining the extension debug verbosity level */
verbosity:		VERBOSITY '=' INTEGER ';'
			{
				rtd_conf->verbosity = $3;
			}
			;
			
	/* Overwrite some default configuration values */
confhdl:		HANDLE_DESTINATION_HOST '=' INTEGER ';'
			{
				rtd_conf->do_dest_host = $3;
			}
			| HANDLE_DESTINATION_REALM '=' INTEGER ';'
			{
				rtd_conf->do_dest_realm = $3;
			}
			;

confscore:		SCORE_DESTINATION_HOST '=' INTEGER ';'
			{
				rtd_conf->sc_dest_host = $3;
			}
			| SCORE_DESTINATION_REALM '=' INTEGER ';'
			{
				rtd_conf->sc_dest_realm = $3;
			}
			| SCORE_NO_INFO '=' INTEGER ';'
			{
				rtd_conf->sc_no_info = $3;
			}
			| SCORE_DEFAULT_PEER '=' INTEGER ';'
			{
				rtd_conf->sc_all = $3;
			}
			| SCORE_DEFAULT_PEER_PREF '=' INTEGER ';'
			{
				rtd_conf->sc_all_pref = $3;
			}
			| SCORE_REALM '=' INTEGER ';'
			{
				rtd_conf->sc_rlm = $3;
			}
			| SCORE_REALM_PREF '=' INTEGER ';'
			{
				rtd_conf->sc_rlm_pref = $3;
			}
			;

	/* Definitions for the routing table */
peerdef:		ROUTE '=' '>' DEST ';'
			{
				if (rtd_add( $1, $4, pref ) != 0) {
					yyerror (&yylloc, conffile, "An error occurred while adding a route definition, aborting...");
					YYERROR;
				}
				/* We don't free the strings here, they will be freed by the rtd_fini function */
			}
			;
			
ROUTE:			REALM
			{
				$$ = $1;
			}
			| '*'
			{
				$$ = NULL;
			}
			;
			
DEST:			FQDN
			{
				pref = 0;
				$$ = $1;
			}
			| FQDN '*'
			{
				pref = 1;
				$$ = $1;
			}
			;

	/* Validating a FQDN: */
FQDN:			QSTRING
			{
				/* Verify this is a valid FQDN */
				if (!is_valid_fqdn($1)) {
					yyerror (&yylloc, conffile, "An error was detected on a fqdn, aborting...");
					YYERROR;
				}
				$$ = $1;
			}
			;
			
	/* Validating a REALM: */
REALM:			QSTRING
			{
				/* Verify this is a valid FQDN */
				if (!is_valid_realm($1)) {
					yyerror (&yylloc, conffile, "An error was detected on a fqdn, aborting...");
					YYERROR;
				}
				$$ = $1;
			}
			;
	
