/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the code to open the listening ports and accept new incoming connections.
 *
 */

#include "waaad-internal.h"
#include "peer-internal.h"

/* Length of server or client thread name */
#define DEBUG_NAME_LEN	40

#define THREAD_NAME_SERVER( servinfo ) \
	log_set_thread_name(  &((_peer_servinfo_t *)(servinfo))->servname[0], NULL  );

#define THREAD_NAME_CLIENT( servinfo, cliinfo ) \
	log_set_thread_name(  	&(((_peer_servinfo_t *)(servinfo))->servname[0]), \
				&(((_peer_cliinfo_t *)(cliinfo))->cliname[0])   );

/* To store a socket and protocol information for the listening thread */
typedef struct {
	int		proto;	/* IPPROTO_TCP or IPPROTO_SCTP */
	int		sock;	/* the socket identifier */
	pthread_t	listen;	/* the thread listening on this socket */
	int		started;/* 1 the thread is started, 2 the thread already exited */
	uti_list_t	clients;/* List of clients informations. the o field of the semaphore points to the head of this object. */
	pthread_mutex_t	climtx;	/* mutex protecting this list */
	char		servname[DEBUG_NAME_LEN+1];/* string for debug */
} _peer_servinfo_t;

/* Store a new client information, before its diameter-id is known and it can be handled properly */
typedef struct {
	uti_list_t	list;	/* link to the list of pending clients. the o fields points to the mutex */
	pthread_t	thr;	/* the thread handling this client */
	int		proto;	/* IPPROTO_TCP or IPPROTO_SCTP */
	int		sock;	/* the socket identifier */
	uint16_t	ostr;	/* outbound streams, for SCTP */
	uint16_t	istr;	/* inbound streams, for SCTP */
	char		cliname[DEBUG_NAME_LEN+1];/* string for debug */
} _peer_cliinfo_t;

#define CI_MTX( _ci_ ) ((pthread_mutex_t *)((_ci_)->list.o))

/* Array of server sockets */
static _peer_servinfo_t * _psi = NULL;
static int _psi_cnt = 0;

/* Cancelation cleanup handler for cleaning a receiver thread */
static void cleanup_peer (void * arg)
{
	_peer_t * dum = _P(arg);

	TRACE_DEBUG(FULL, "Cleaning up receiver thread for dummy peer %p", arg);
	
	CHECK_FCT_DO(  _peer_struct_destroy(&dum),  );
	
	/* End of cleanup handler */
	return;
}

static void cleanup_lock (void * arg)
{
	CHECK_POSIX_DO(  pthread_mutex_unlock(arg),  /* nothing */ );
	return;
}

static void cleanup_pmsg (void * arg)
{
	msg_t ** pmsg = (msg_t **)arg;
	if (pmsg && *pmsg) {
		CHECK_FCT_DO(  msg_free(*pmsg, 1),  /* nothing */ );
	}
	return;
}


/* The code of a receiver thread, to handle an incoming connection, before the CER is received */
static void * _peer_handle_client_th(void * arg)
{
	_peer_cliinfo_t *ci = (_peer_cliinfo_t *) arg;
	_peer_t * dummy;
	struct timespec	delay;
	_pe_t * event;
	msg_t * received = NULL;
	uint32_t ete, hbh;
	
	THREAD_NAME_CLIENT( ci->list.head->o, ci );
	
	TRACE_ENTRY( "%p", arg );
	
	TRACE_DEBUG(FULL, "Thread created to handle a new client: %p", pthread_self());
	
	/* let's create a dummy peer to store all CER data */
	CHECK_FCT_DO( _peer_struct_malloc(NULL, &dummy), goto error  );
	pthread_cleanup_push(cleanup_peer, dummy);
	
	/* we'll use this information later */
	dummy->p_addinfo.pa_proto = ci->proto;
	dummy->p_sock = ci->sock;
	ci->sock = 0; /* prevent closing the socket on cleanup */
	dummy->p_ostr = ci->ostr;
	dummy->p_istr = ci->istr;
	
	/* the sec_session_t object used by the security module */
	SEC_SESS_INIT( &dummy->p_sec_session, ci->proto, dummy, &dummy->p_sock );
	
	/* Lock the peer to wait for incoming events */
	CHECK_POSIX_DO(  pthread_mutex_lock(&dummy->p_lock),  goto error  );
	pthread_cleanup_push(cleanup_lock, &dummy->p_lock);
	
	/* Create the receiver thread */
	CHECK_FCT_DO(  _peer_in_start(dummy),  goto error  );

	/* Compute the timeout absolute time for receiving the CER message */
	CHECK_SYS_DO(  clock_gettime(CLOCK_REALTIME, &delay),  goto error  );
	delay.tv_sec += INCNX_TIMEOUT;
	
	/* Stack the cleanup handler, in case we are canceled */
	pthread_cleanup_push(cleanup_pmsg, &received);
			
	
	/* Now wait for an event to occur. Such events include: connection broken, message received. 
	 In any case, if we have not received a CER message, we just abort the connection and free the resources.
	 -- should we send an error message? will be implemented later eventually --
	 */
spurious:
	CHECK_POSIX_DO2( pthread_cond_timedwait( &dummy->p_condvar, &dummy->p_lock, &delay ),
		ETIMEDOUT,  {
				log_error("No CER received on incoming connection within %d seconds, closing the connection.\n", INCNX_TIMEOUT);
				goto error;
			},
		goto error );

	/* Check for spurious wakeup */
	if (IS_LIST_EMPTY( &dummy->p_events ))
		goto spurious;
	
	/* Ok, retrieve the next event */
	event = (_pe_t *)(dummy->p_events.next);
	uti_list_unlink( _LIST(event) );
	
	TRACE_DEBUG(FULL, "Thread %p received event %s", pthread_self(), EVENTSTR(event->event));
	
	switch (event->event) {
		/* The following events are ignored */
		case PEVENT_ASSOC_CHG:
		case PEVENT_PEERADDR_CHG:
			goto spurious; 
		
		/* This is the event we are wanting */
		case PEVENT_MSGRCVD:
			/* Retrieve the message from the queue */
			CHECK_FCT_DO(  meq_tryget(dummy->p_in_q, &received),  goto error  );
			
			/* Parse the message against the dictionary */
			CHECK_FCT_DO(  msg_parse_dict(received), goto no_cer  );
				
			/* Check we have received a CER */
			{
				msg_data_t *mdata;
				
				CHECK_FCT_DO(  msg_data( received, &mdata ), goto no_cer  );
				
				if (mdata->msg_appl != 0) {
					log_error("First message received on incoming connection is not Diameter Base Protocol (%u), closing...\n", mdata->msg_appl);
					goto no_cer;
				}
				if (mdata->msg_code != CC_CAPABILITIES_EXCHANGE) {
					log_error("First message received on incoming connection is not a CER (%u), closing...\n", mdata->msg_code);
					goto no_cer;
				}
				
				/* Now check we support all mandatory AVPs */
				CHECK_FCT_DO(  msg_parse_rules(received, NULL), goto no_cer  );
				
				/* Save the hop-by-hop and end-to-end for later */
				hbh = mdata->msg_hbhid;
				ete = mdata->msg_eteid;
			}
			
			/* We have received a valid CER */
			break;
		
		/* Other events are considered as errors */
		default:
			goto error;
	}
	
	/* We received a CER */
	
	/* Cancel receiver thread */
	CHECK_FCT_DO(  _peer_in_stop(dummy), goto error  );
	
	/* Now parse the content of the received CER */
	CHECK_FCT_DO(  _peer_cer_parse( received, dummy ), goto error  );
	
	/* From this point we can discard the CER, we won't need it anymore (hbhid & eteid already saved) */
	CHECK_FCT_DO(  msg_free(received, 1), /* nothing */ );
	received = NULL;
	
	/* Save also the connection information in the peer structure */
	{
		socklen_t len = sizeof(dummy->p_addinfo.pa_ss);
		CHECK_SYS_DO( getpeername( dummy->p_sock, (sSA *)&dummy->p_addinfo.pa_ss, &len ), goto error );
	}
	
	/* The dummy structure contains all parsed CER data */
	{
		uti_list_t * next = NULL;
		int ret = _peer_struct_search(dummy->p_diamid, 0, &dummy->p_hash, &next);
		/* Did we find the peer? */
		switch (ret) {
			case 0:
				pthread_cleanup_push(cleanup_lock, H_LOCK( dummy->p_hash ));
				{
					/* the peer already exists, send it an event, the PSM will deal with incoming connection depending on its state */
					_ev_incoming_t * ev;
					_peer_t * ref = _P(next->o);

					/* Lock the peer */
					CHECK_POSIX_DO( pthread_mutex_lock( &ref->p_lock ), 
							{ cleanup_lock(H_LOCK( dummy->p_hash )); goto error; } );
					
					/* Check its state */
					if ( ref->p_state == STATE_DISABLED ) {
						 /* Note: in case the peer is in DISABLED state, what do we do? */
						 /* -> start the PSM ? */
						 /* -> reject the CER with an error? */
						 /* It depends why the peer is in DISABLED state... we have to check this */
						TRACE_DEBUG(INFO, "@@@ Not implemented yet");
						
						/* If the peer state machine must not be started for some reason, reject the CER (error msg) and close */
						ASSERT(0);
						
						/* otherwise, simply start the PSM then send an event */
						CHECK_FCT_DO(  _peer_psm_start( ref, 2),
							{ 
								CHECK_POSIX_DO( pthread_mutex_unlock( &ref->p_lock ), /* ignore */ );
								cleanup_lock(H_LOCK( dummy->p_hash )); 
								goto error; 
							} );
					}
					/* Ok, the PSM is running, just send an event */

					CHECK_MALLOC_DO( ev = (_ev_incoming_t *) malloc(sizeof(_ev_incoming_t)),
						{ 
							CHECK_POSIX_DO( pthread_mutex_unlock( &ref->p_lock ), /* ignore */ );
							cleanup_lock(H_LOCK( dummy->p_hash )); 
							goto error; 
						} );

					memset(ev, 0, sizeof(_ev_incoming_t));
					ev->peer = dummy;
					ev->hbh = hbh;
					ev->ete = ete;

					CHECK_FCT_DO( _peer_events_send_nl( ref, PEVENT_INCNX, ev ),
						{ 
							free(ev);
							CHECK_POSIX_DO( pthread_mutex_unlock( &ref->p_lock ), /* ignore */ );
							cleanup_lock(H_LOCK( dummy->p_hash )); 
							goto error; 
						} );

					/* Ok, we are done */
					CHECK_POSIX_DO(  pthread_mutex_unlock(&dummy->p_lock), goto error  );
					dummy = NULL;
						
					CHECK_POSIX_DO( pthread_mutex_unlock( &ref->p_lock ), 
							{ cleanup_lock(H_LOCK( dummy->p_hash )); goto error; } );
						
				}
				pthread_cleanup_pop(1);
				break;
			
			case ENOENT:
				pthread_cleanup_push(cleanup_lock, H_LOCK( dummy->p_hash ));
				/* This is a new peer, we must decide if we can handle it, then create a PSM for it. */
				/*
				   CERs received from unknown peers MAY be silently discarded, or a CEA
				   MAY be issued with the Result-Code AVP set to DIAMETER_UNKNOWN_PEER.
				   In both cases, the transport connection is closed.  If the local
				   policy permits receiving CERs from unknown hosts, a successful CEA
				   MAY be returned.  If a CER from an unknown peer is answered with a
				   successful CEA, the lifetime of the peer entry is equal to the
				   lifetime of the transport connection.  In case of a transport
				   failure, all the pending transactions destined to the unknown peer
				   can be discarded.
				*/				
				TRACE_DEBUG (INFO, "@@@: support for unknown peers not implemented yet.");
				
				/* Currently, we just reject this connection attempt */
				
				pthread_cleanup_pop(1); /* temporary, we will pop(0) later */
				goto error;
			
			default:
				/* An unexpected error occurred -- we suppose the line was not locked */
				CHECK_FCT_DO( ret, goto error ); /* this will always go to error after logging the error */
		}
		
		/* We are done */
		goto end;	
	}
	
no_cer:	
	/* We received an invalid event or message */
	
	/* Reply with an error code? */

error:	/* Destroy connection information and threads */
	
	/* Destroy the message as needed */
	if (received) {
		CHECK_FCT_DO(  msg_free(received, 1), /* nothing */ );
	}
	pthread_cleanup_pop(0); /* cleanup pmsg */
	
	if (dummy) {
		CHECK_POSIX_DO(  pthread_mutex_unlock(&dummy->p_lock),  /* nothing */  );
	}
	pthread_cleanup_pop(0); /* cleanup unlock */
	
	/* Destroy the dummy peer */
	if (dummy) {
		CHECK_FCT_DO(  _peer_struct_destroy(&dummy),  );
	}
	pthread_cleanup_pop(0);
	
	/* close the socket */
	if (ci->sock)
		shutdown(ci->sock, SHUT_RDWR);

end:	
	/* Remove the client information from the list */
	CHECK_POSIX_DO(  pthread_mutex_lock(CI_MTX(ci)), /* nothing */  );
	
	uti_list_unlink(_LIST(ci));
	
	CHECK_POSIX_DO(  pthread_mutex_unlock(CI_MTX(ci)), /* nothing */  );
	
	/* detach this thread */
	pthread_detach(pthread_self());
	
	/* free this client resource */
	free(ci);
	
	/* Exit */
	return NULL;
}

/* The code of the thread waiting for incoming connection. 1 thread per listening socket. */
static void * _peer_server_th(void * arg)
{
	_peer_servinfo_t * si = (_peer_servinfo_t *)arg;
	_peer_cliinfo_t  * ci = NULL;
	
	THREAD_NAME_SERVER( si );
	
	TRACE_ENTRY( "%p", arg );
	
	si->started = 1;
	
	/* Initialize the list of clients */
	uti_list_init ( &si->clients, si );
	CHECK_POSIX_DO(  pthread_mutex_init(&si->climtx, NULL),  goto error  );
	
	/* Listen, to enable remote peers to connect */
	CHECK_SYS_DO(  listen(si->sock, 5), { TRACE_DEBUG(FULL, "Error on socket %d", si->sock); goto error; }  );
	
	/* loop */
	while (1) {
		/* Create a cliinfo object to receive new client connection information */
		CHECK_MALLOC_DO(  ci = (_peer_cliinfo_t *) malloc(sizeof(_peer_cliinfo_t)),  goto error  );
		memset(ci, 0, sizeof(_peer_cliinfo_t));
		uti_list_init ( &ci->list, &si->climtx );
		ci->proto = si->proto;
		
		/* Accept a new client */
		switch (si->proto) {
			case IPPROTO_TCP:
				/* we use a simple accept */
				CHECK_SYS_DO(  ci->sock = accept(si->sock, NULL, NULL),  goto error  );
				break;
			
			case IPPROTO_SCTP:
				/* We use our wrapper */
				CHECK_FCT_DO(  _peer_sctp_accept( si->sock, &ci->sock, &ci->ostr, &ci->istr ),  goto error  );
				break;
			
			default:
				assert(0);
		}
		snprintf( &(ci->cliname[0]), DEBUG_NAME_LEN, "Cli(%d)", ci->sock);
	
		/* Save this new client in the list of the server */
		CHECK_POSIX_DO(  pthread_mutex_lock(&si->climtx),  goto error  );
		
		uti_list_insert_before ( &si->clients, &ci->list );

		CHECK_POSIX_DO(  pthread_mutex_unlock(&si->climtx),  goto error  );
		
		/* And now start the received thread for this client */
		CHECK_POSIX_DO(  pthread_create( &ci->thr, NULL, _peer_handle_client_th, ci),  goto error  );
	}
		
error:
	TRACE_DEBUG(INFO, " Error occurred in listening server thread");
	TRACE_DEBUG(INFO, "  Proto : %s", (si->proto == IPPROTO_TCP) ? "IPPROTO_TCP" : 
						(si->proto == IPPROTO_SCTP) ? "IPPROTO_SCTP" : 
							"???");
	TRACE_DEBUG(INFO, "  Sock  : %d", si->sock);

	si->started = 2;

	/* Signal the daemon that something went wrong here ... */
	CHECK_FCT_DO(  uti_event_send(WDE_ERROR),  /* nothing */  );	

	return NULL;
}

/* Count the number of servers that will be created, according to the configuration */
static int _count_tcp_servers() 
{
	sSS * ss = g_pconf->local_addr_pri_sa;
	int count = 0;
	
	/* if disable_tcp => 0 */
	if (g_pconf->disable_tcp)
		return 0;
	
	/* if local_addr_pri_sa is NULL => 1 or 2 -- will bind to ANY */
	if (ss == NULL) {
		if (g_pconf->disable_inet4 || g_pconf->disable_inet6)
			return 1;
		else
			return 2;
	}
	
	/* otherwise count the number of elements in the list and return it */
	while (ss->ss_family != 0) {
		count++;
		ss++;
	}
	
	return count;
}

/* Code to start listening */
int _peer_server_start()
{
	int i   = 1;
	
	TRACE_ENTRY( );
	
	/* First, count the number of sockets we will have to create, in addition to the SCTP one */
	_psi_cnt = 1 /* sctp */ + _count_tcp_servers();
	
	/* Allocate the _psi array */
	CHECK_MALLOC( _psi = calloc(_psi_cnt, sizeof(_peer_servinfo_t)) );
	
	/* The first element is the SCTP server, create this one already */
	_psi[0].proto = IPPROTO_SCTP;
	CHECK_FCT(  _peer_sctp_create_server( &_psi[0].sock )  );
	snprintf( &(_psi[0].servname[0]), DEBUG_NAME_LEN, "SCTP.Srv(%d)", _psi[0].sock);
	
	/* Now create the TCP servers */
	if ( ! g_pconf->disable_tcp ) {
		sSS stor;
		sSA  * sa  = (sSA  *)&stor;
		sSA4 * sin = (sSA4 *)&stor;
		sSA6 * sin6 = (sSA6 *)&stor;
		sSS  * ss = g_pconf->local_addr_pri_sa;
		
		if (ss == NULL) {
			/* The list of addresses is not specified, bind to any */
			
			if (! g_pconf->disable_inet4 ) {
				
				memset(&stor, 0, sizeof(sSS));
				sin->sin_family = AF_INET;
				sin->sin_port = htons(g_pconf->local_port);
				sin->sin_addr.s_addr = INADDR_ANY;
				
				_psi[i].proto = IPPROTO_TCP;
				CHECK_FCT(  _peer_tcp_create_server( &(_psi[i].sock), sa, sizeof(sSA4) )  );
				
				snprintf( &(_psi[i].servname[0]), DEBUG_NAME_LEN, "TCP.Srv(%d)(0.0.0.0:%hu)", _psi[i].sock, g_pconf->local_port);
				i++;
			}
					
			if (! g_pconf->disable_inet6 ) {
				
				memset(&stor, 0, sizeof(sSS));
				sin6->sin6_family = AF_INET6;
				sin6->sin6_port = htons(g_pconf->local_port);
				sin6->sin6_addr = in6addr_any;
				
				_psi[i].proto = IPPROTO_TCP;
				CHECK_FCT(  _peer_tcp_create_server( &(_psi[i].sock), sa, sizeof(sSA6) )  );
				
				snprintf( &(_psi[i].servname[0]), DEBUG_NAME_LEN, "TCP.Srv(%d)([::]:%hu)", _psi[i].sock, g_pconf->local_port);
				i++;
			}
		} else {
			/* The list of addresses to bind to is in the configuration file */
			while (ss->ss_family != 0) {

				_psi[i].proto = IPPROTO_TCP;
				CHECK_FCT(  _peer_tcp_create_server( &(_psi[i].sock), (sSA *)ss, sizeof(sSS) )  );
				
				{
					char __addrbuf[INET6_ADDRSTRLEN];
					switch (ss->ss_family) {
						case AF_INET:
							snprintf( &(_psi[i].servname[0]), DEBUG_NAME_LEN, "TCP.Srv(%d)(%s:%hu)", _psi[i].sock, 
									inet_ntop(ss->ss_family, &((sSA4 *)ss)->sin_addr.s_addr, __addrbuf, INET6_ADDRSTRLEN),
									g_pconf->local_port);
							break;
						case AF_INET6:
							snprintf( &(_psi[i].servname[0]), DEBUG_NAME_LEN, "TCP.Srv(%d)(%s:%hu)", _psi[i].sock, 
									inet_ntop(ss->ss_family, &((sSA6 *)ss)->sin6_addr.s6_addr, __addrbuf, INET6_ADDRSTRLEN),
									g_pconf->local_port);
							break;
						default:
							snprintf( &(_psi[i].servname[0]), DEBUG_NAME_LEN, "TCP.Srv(%d)(???:%hu)", _psi[i].sock, g_pconf->local_port);
					}
				}
				ss++;
				i++;
			}
		}
		
		ASSERT(i == _psi_cnt);
	}
	
	TRACE_DEBUG(FULL, "All %d server sockets are created, now starting the server threads...", _psi_cnt);
	
	/* Start all the _peer_listen_th threads */
	for (i=0; i < _psi_cnt; i++) {
		CHECK_POSIX(  pthread_create( &_psi[i].listen, NULL, _peer_server_th, (void *)&_psi[i] )  );
	}
	
	return 0;
}

int _peer_server_stop()
{
	int i = 0;
	
	TRACE_ENTRY( );
	
	if (_psi) {
		/* loop in the list, kill the threads (clients and server) if any, close the sockets */
		for (i = 0; i < _psi_cnt; i++) {
			
			/* First, stop the server thread so that no new client is accepted */
			CHECK_FCT_DO( _thread_term( &(_psi[i].listen) ), /* continue */ );

			/* Lock the client list */
			CHECK_POSIX_DO(  pthread_mutex_lock( &(_psi[i].climtx) ), /* continue */ );
			
			/* Kill any client thread */
			while ( !IS_LIST_EMPTY( &(_psi[i].clients) )  ) {
				_peer_cliinfo_t * cli = (_peer_cliinfo_t *) (_psi[i].clients.next);
				
				/* unlink */
				uti_list_unlink(_LIST(cli));
				
				/* Kill the thread */
				CHECK_FCT_DO( _thread_term( &cli->thr ),  /* continue */);
				
				/* Close the socket */
				CHECK_SYS_DO(  shutdown(cli->sock, SHUT_RDWR),  /* continue */);
				
				/* free this client information */
				free(cli);
			}
			
			/* Unlock and destroy the client mutex */
			CHECK_POSIX_DO(  pthread_mutex_unlock( &(_psi[i].climtx) ), /* continue */ );
			CHECK_POSIX_DO(  pthread_mutex_destroy( &(_psi[i].climtx) ), /* continue */ );
			
			/* Now shutdown the server socket */
			CHECK_SYS_DO(  shutdown(_psi[i].sock, SHUT_RDWR),  /* continue */);
		}
		
		/* then free the object */
		free(_psi);
		_psi = NULL;
	}
	
	return 0;
}
