/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

#include "tests.h"
#include <netinet/in.h>
#include <arpa/inet.h>

extern void dump_dict_object ( void * obj, int parents, int depth, int indent );


/* Main test routine */
int main(int argc, char *argv[])
{
	msg_t * acr = NULL;
	msg_avp_t * pi = NULL, *avp1, *avp2;
	unsigned char * buf = NULL;
	
	/* First, initialize the daemon modules */
	INIT_WAAAD();
	
	/* Create the message object from model */
	{
		dict_object_t * acr_model = NULL;

		/* Now find the ACR dictionary object */
		CHECK( 0, dict_search ( DICT_COMMAND, CMD_BY_NAME, "Accounting-Request", &acr_model ) );

		/* Create the instance, using the templates */
		CHECK( 0, msg_new ( acr_model, MSGFL_FROM_TEMPLATE, &acr ) );

		#if 0
		/* For debug: dump the object */
		log_debug("Dumping Accounting-Request message (from template)\n");
		msg_dump_walk(acr);
		#endif
	}
	
	/* Create the Proxy-Info AVP from model */
	{
		dict_object_t * pi_model = NULL;

		/* Now find the ACR dictionary object */
		CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME, "Proxy-Info", &pi_model ) );

		/* Create the instance, using the templates */
		CHECK( 0, msg_avp_new ( pi_model, MSGFL_FROM_TEMPLATE, &pi ) );

		#if 0
		/* For debug: dump the object */
		log_debug("Dumping Proxy-Info AVP (from template)\n");
		msg_dump_walk(pi);
		log_debug("Dumping dictionary model\n");
		dict_dump(pi_model, 0, 2);
		#endif
		
		/* Check that msg_free does not recursively delete if not told to. */
		CHECK( EINVAL, msg_free( pi, 0 ) );
	}
	
	/* Get a reference to the current last AVP in the message */
	{
		int diff = 0;
		CHECK( 0, msg_browse ( acr, MSG_BRW_LAST_CHILD, &avp1, &diff) );
		CHECK( 1, diff );
		
		/* Check that we cannot add this AVP to another message since it is already linked */
		CHECK( EINVAL, msg_avp_add( pi, MSG_BRW_LAST_CHILD, avp1) );
	}

	/* Now add the Proxy-Info AVP at the end of the message */
	{
		CHECK( 0, msg_avp_add( acr, MSG_BRW_LAST_CHILD, pi) );
		#if 0
		/* For debug: dump the object */
		log_debug("Dumping Accounting-Request with Proxy-Info AVP at the end\n");
		msg_dump_walk(acr);
		#endif
	}
	
	/* Check the last child is now the proxy-Info */
	{
		CHECK( 0, msg_browse ( acr, MSG_BRW_LAST_CHILD, &avp2, NULL) );
		CHECK( pi, avp2 );
	}
	
	/* Check that the avp before the proxy-info is the previous last one */
	{
		int diff = 0;
		CHECK( 0, msg_browse ( pi, MSG_BRW_PREV, &avp2, &diff) );
		CHECK( avp1, avp2 );
		CHECK( 0, diff);
	}
	
	/* Check that there are no AVP after the proxy-info */
	CHECK( ENOENT, msg_browse ( pi, MSG_BRW_NEXT, NULL, NULL) );
	
	/* Test the msg_avp_del function unlinks the object properly */
	{
		dict_object_t * rr_model = NULL;

		/* Now find the ACR dictionary object */
		CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME, "Route-Record", &rr_model ) );

		/* Create the instance, using the templates */
		CHECK( 0, msg_avp_new ( rr_model, MSGFL_FROM_TEMPLATE, &avp1 ) );
		
		/* Add the AVP at the end of the message */
		CHECK( 0, msg_avp_add( pi, MSG_BRW_NEXT, avp1) );
		
		/* Check the last AVP of the message is now this one */
		CHECK( 0, msg_browse ( acr, MSG_BRW_LAST_CHILD, &avp2, NULL) );
		CHECK( avp1, avp2 );
		
		/* Now delete it */
		CHECK( 0, msg_free( avp1, 0 ) );
		
		/* Check the last AVP of the message is back to pi */
		CHECK( 0, msg_browse ( acr, MSG_BRW_LAST_CHILD, &avp2, NULL) );
		CHECK( pi, avp2 );
		
		/* Delete the whole message */
		CHECK( 0, msg_free( acr, 1 ) );
	}
	
	/* Recreate the ACR without any AVP */
	{
		dict_object_t * acr_model = NULL;

		/* Now find the ACR dictionary object */
		CHECK( 0, dict_search ( DICT_COMMAND, CMD_BY_NAME, "Accounting-Request", &acr_model ) );

		/* Create the instance, using the templates */
		CHECK( 0, msg_new ( acr_model, 0, &acr ) );

		#if 0
		/* For debug: dump the object */
		log_debug("Dumping Accounting-Request message (from template)\n");
		msg_dump_walk(acr);
		#endif
		
		/* Check there is no child */
		CHECK( ENOENT, msg_browse ( acr, MSG_BRW_FIRST_CHILD, NULL, NULL) );
	}
	
	/* Now let's create some additional Dictionary objects for the test */
	{
		/* The constant values used here are totally arbitrary chosen */
		dict_object_t * vendor;
		{
			dict_vendor_data_t vendor_data = { 73565, "Vendor test" };
			CHECK( 0, dict_new ( DICT_VENDOR, &vendor_data , NULL, &vendor ) );
		}
		
		{
			dict_application_data_t app_data = { 73566, "Application test" };
			CHECK( 0, dict_new ( DICT_APPLICATION, &app_data , vendor, NULL ) );
		}
		
		{
			dict_avp_data_t avp_data = { 73567, 0, "AVP Test - no vendor - f32", 0, 0, AVP_TYPE_FLOAT32 };
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , NULL, NULL ) );
		}
		
		{ 
			dict_object_t * type = NULL;
			dict_type_data_t type_data = { AVP_TYPE_INTEGER64, "Int64 test" };
			dict_avp_data_t avp_data = { 73568, 73565, "AVP Test - i64", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_INTEGER64 };
			CHECK( 0, dict_new ( DICT_TYPE, &type_data , NULL, &type ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , type, NULL ) );
		}
		
		{
			dict_object_t * type = NULL;
			dict_type_data_t type_data = { AVP_TYPE_INTEGER32, "Enum32 test" };
			dict_type_enum_data_t val1 = { "i32 const test (val 1)", { .i32 = 1 } };
			dict_type_enum_data_t val2 = { "i32 const test (val 2)", { .i32 = 2 } };
			dict_type_enum_data_t val3 = { "i32 const test (val -5)",{ .i32 = -5 } };
			dict_avp_data_t avp_data = { 73569, 73565, "AVP Test - enumi32", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_INTEGER32 };
			
			CHECK( 0, dict_new ( DICT_TYPE, &type_data , NULL, &type ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val1 , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val2 , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val3 , type, NULL ) );
		}
			
		{ 
			dict_object_t * type = NULL;
			dict_type_data_t type_data = { AVP_TYPE_OCTETSTRING, "OS test" };
			dict_avp_data_t avp_data = { 73570, 73565, "AVP Test - os", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_OCTETSTRING };
			CHECK( 0, dict_new ( DICT_TYPE, &type_data , NULL, &type ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , type, NULL ) );
		}
		
		{
			dict_object_t * type = NULL;
			dict_type_data_t type_data = { AVP_TYPE_OCTETSTRING, "OS enum test" };
			dict_type_enum_data_t val1 = { "os const test (Test)", { .os = { (unsigned char *)"Test", 4 } } };
			dict_type_enum_data_t val2 = { "os const test (waaad)", { .os = { (unsigned char *)"waaad", 5 } } };
			dict_type_enum_data_t val3 = { "os const test (waa)", { .os = { (unsigned char *)"waaad", 3 } } };
			dict_avp_data_t avp_data = { 73571, 73565, "AVP Test - enumos", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_OCTETSTRING };
			
			CHECK( 0, dict_new ( DICT_TYPE, &type_data , NULL, &type ) );
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val1 , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val2 , type, NULL ) );
			CHECK( 0, dict_new ( DICT_TYPE_ENUM, &val3 , type, NULL ) );
		}
		
		{
			dict_object_t * gavp = NULL;
			dict_avp_data_t avp_data = { 73572, 73565, "AVP Test - grouped", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_GROUPED };
			
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , NULL, &gavp ) );
			
			/* Macro to search AVP and create a rule */		
			#define ADD_RULE( _parent, _vendor, _avpname, _pos, _min, _max, _tpl, _ord ) {		\
				dict_object_t * _avp = NULL;							\
				dict_avp_request_t _req = { (_vendor), 0, (_avpname) };				\
				dict_rule_data_t _data;								\
				CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME_AND_VENDOR, &_req, &_avp ) );	\
				_data.rule_avp = _avp;								\
				_data.rule_position = (_pos);							\
				_data.rule_order = (_ord);							\
				_data.rule_min = (_min);							\
				_data.rule_max = (_max);							\
				_data.rule_template = (_tpl);							\
				CHECK( 0, dict_new ( DICT_RULE, &_data , (_parent), NULL ) );			\
			}
			
			ADD_RULE(gavp, 73565, "AVP Test - os", RULE_OPTIONAL,   -1, -1,  3, 0);
			
		}
			
		{
			dict_object_t * application = NULL;
			dict_object_t * command = NULL;
			dict_cmd_data_t cmd_data = { 73573, "Test-Command-Request", CMD_FLAG_REQUEST, CMD_FLAG_REQUEST };
			
			CHECK( 0, dict_search ( DICT_APPLICATION, APPLICATION_BY_NAME, "Application test", &application ) );
			CHECK( 0, dict_new ( DICT_COMMAND, &cmd_data , application, &command ) );
			ADD_RULE(command, 0,     "AVP Test - no vendor - f32", 	RULE_FIXED_HEAD, -1,  1,  0, 1);
			ADD_RULE(command, 73565, "AVP Test - i64",		RULE_REQUIRED,   -1, -1,  0, 0);
			ADD_RULE(command, 73565, "AVP Test - enumi32", 		RULE_OPTIONAL,   -1, -1,  0, 0);
			ADD_RULE(command, 73565, "AVP Test - os", 		RULE_OPTIONAL,   -1, -1,  0, 0);
			ADD_RULE(command, 73565, "AVP Test - enumos", 		RULE_OPTIONAL,   -1, -1,  0, 0);
			ADD_RULE(command, 73565, "AVP Test - grouped", 		RULE_OPTIONAL,   -1, -1,  0, 0);
		}
		
		{
			dict_object_t * gavp = NULL;
			dict_avp_data_t avp_data = { 73574, 73565, "AVP Test - rules", AVP_FLAG_VENDOR, AVP_FLAG_VENDOR, AVP_TYPE_GROUPED };
			
			CHECK( 0, dict_new ( DICT_AVP, &avp_data , NULL, &gavp ) );
			
			ADD_RULE(gavp,     0, "AVP Test - no vendor - f32", RULE_FIXED_HEAD,   0, 1, 0, 1);
			ADD_RULE(gavp, 73565, "AVP Test - i64", 	    RULE_FIXED_HEAD,  -1, 1, 0, 2);
			ADD_RULE(gavp, 73565, "AVP Test - enumi32", 	    RULE_FIXED_HEAD,  -1, 1, 0, 3);
			ADD_RULE(gavp, 73565, "AVP Test - os", 	    	    RULE_REQUIRED,     2, 3, 0, 0);
			ADD_RULE(gavp, 73565, "AVP Test - enumos",     	    RULE_OPTIONAL,     0, 1, 0, 0);
			ADD_RULE(gavp, 73565, "AVP Test - grouped",         RULE_FIXED_TAIL,  -1, 1, 0, 1);
			
			#if 0
			dict_dump ( gavp, 0, 2 );
			#endif
		}
		#if 0
		{
			dict_dump ( vendor, 1, 2 );
		}
		#endif
	}
	
	/* Now create some values and check the length is correctly handled */
	{
		dict_object_t * cmd_model = NULL;
		msg_t         * msg = NULL;
		dict_object_t * avp_model = NULL;
		msg_avp_t     * avp = NULL;
		avp_value_t	value;
		
		CHECK( 0, dict_search ( DICT_COMMAND, CMD_BY_NAME, "Test-Command-Request", &cmd_model ) );
		
		/* Check an error is trigged if the AVP has no value set */
		{
			CHECK( 0, dict_search ( DICT_AVP,     AVP_BY_NAME,     "AVP Test - no vendor - f32", &avp_model ) );
			
			CHECK( 0, msg_new ( cmd_model, 0, &msg ) );
			CHECK( 0, msg_avp_new ( avp_model, 0, &avp ) );
			
			CHECK( 0, msg_avp_add ( msg, MSG_BRW_FIRST_CHILD, avp ) );
			
			CHECK( EINVAL, msg_update_length ( avp ) );
			CHECK( EINVAL, msg_update_length ( msg ) );
			
			CHECK( 0, msg_free( msg, 1 ) );
		}
		
		/* Check the sizes are handled properly */
		{
			msg_avp_t * avpi = NULL;
			msg_avp_t * avpch = NULL;
			msg_avp_data_t * avpdata = NULL;
			msg_data_t * msgdata = NULL;
			#define ADD_AVP( _parent, _position, _avpi, _avpvendor, _avpname) {			\
				dict_object_t * _avp = NULL;							\
				dict_avp_request_t _req = { (_avpvendor), 0, (_avpname) };			\
				CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME_AND_VENDOR, &_req, &_avp ) );	\
				CHECK( 0, msg_avp_new ( _avp, 0, &_avpi ) );					\
				CHECK( 0, msg_avp_add ( (_parent), (_position), _avpi ) );			\
			}
			/* Create a message with many AVP inside */
			CHECK( 0, msg_new ( cmd_model, 0, &msg ) );
			CHECK( 0, msg_data ( msg, &msgdata ) );
			
			/* Avp no vendor, float32 => size = 12 */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 0,     "AVP Test - no vendor - f32" );
			value.f32 = 3.1415;
			CHECK( 0, msg_avp_setvalue ( avpi, &value ) );
			CHECK( 0, msg_update_length ( avpi ) );
			#if 0
			log_debug("AVP no vendor, value 3.1415:\n");
			msg_dump_one(avpi);
			#endif
			CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
			CHECK( 12, avpdata->avp_len );
			
			/* Check what happens when we delete the value */
			CHECK( 0, msg_avp_setvalue ( avpi, NULL ) );
			CHECK( EINVAL, msg_update_length ( avpi ) );
			CHECK( 0, msg_avp_setvalue ( avpi, &value ) );
			
			/* Add a vendor AVP, integer64 => size = 20 */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - i64" );
			value.i64 = 0x123456789abcdeLL;
			CHECK( 0, msg_avp_setvalue ( avpi, &value ) );
			CHECK( 0, msg_update_length ( avpi ) );
			#if 0
			log_debug("AVP vendor, value 0x123456789abcdeL:\n");
			msg_dump_one(avpi);
			#endif
			CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
			CHECK( 20, avpdata->avp_len );
			
			/* Check the size of the message is 20 (header) + 12 + 20 = 52 */
			CHECK( 0, msg_update_length ( msg ) );
			CHECK( 52, msgdata->msg_length );
			
			/* Add an AVP with an enum value */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumi32" );
			{
				dict_object_t * type_model = NULL;
				dict_object_t * value_model = NULL;
				dict_type_enum_request_t request;
				
				CHECK( 0, msg_model ( avpi, &avp_model ) );
				CHECK( 0, dict_search ( DICT_TYPE, TYPE_OF_AVP, avp_model, &type_model ) );
				memset(&request, 0, sizeof(request));
				request.type_obj = type_model;
				request.search.enum_name = "i32 const test (val 2)";
				CHECK( 0, dict_search ( DICT_TYPE_ENUM, ENUM_BY_STRUCT, &request, &value_model ) );
				CHECK( 0, dict_getval ( value_model, &request.search ) );
				CHECK( 0, msg_avp_setvalue ( avpi, &request.search.enum_value ) );
				#if 0
				log_debug("AVP enum i32, value 2:\n");
				msg_dump_one(avpi);
				#endif
			}
			
			/* Add an AVP with an enum value, negative */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumi32" );
			{
				dict_object_t * type_model = NULL;
				dict_object_t * value_model = NULL;
				dict_type_enum_request_t request;
				
				CHECK( 0, msg_model ( avpi, &avp_model ) );
				CHECK( 0, dict_search ( DICT_TYPE, TYPE_OF_AVP, avp_model, &type_model ) );
				memset(&request, 0, sizeof(request));
				request.type_obj = type_model;
				request.search.enum_name = "i32 const test (val -5)";
				CHECK( 0, dict_search ( DICT_TYPE_ENUM, ENUM_BY_STRUCT, &request, &value_model ) );
				CHECK( 0, dict_getval ( value_model, &request.search ) );
				CHECK( 0, msg_avp_setvalue ( avpi, &request.search.enum_value ) );
				#if 0
				log_debug("AVP enum i32, value -5:\n");
				msg_dump_one(avpi);
				#endif
				/* Check the size is correct ( 12 for header + 4 for value ) */
				CHECK( 0, msg_update_length ( avpi ) );
				CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
				CHECK( 16, avpdata->avp_len );
			}
			
			/* Now add a value which is not a constant into a enumerated AVP */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumi32" );
			value.i32 = -10;
			CHECK( 0, msg_avp_setvalue ( avpi, &value ) );
			CHECK( 0, msg_update_length ( avpi ) );
			#if 0
			log_debug("AVP vendor enum i32, value -10 (not enum):\n");
			msg_dump_one(avpi);
			#endif
			
			/* Add an octetstring AVP */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - os" );
			{
				unsigned char buf[90];
				memcpy(&buf, "This\0 is a buffer of dat\a. It is not a string so we can have any c\0ntr\0l character here...\0\0", 89);
				value.os.data = buf;
				value.os.len = 89;
				CHECK( 0, msg_avp_setvalue ( avpi, &value ) );
				memset(&buf, 0, sizeof(buf)); /* Test that the OS value is copied */
				CHECK( 0, msg_update_length ( avpi ) );
				#if 0
				log_debug("AVP octet string, 'This\\0 is a b...'\n");
				msg_dump_one(avpi);
				#endif
				CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
				CHECK( 101, avpdata->avp_len );
				CHECK( 'T', avpdata->avp_data->os.data[0] );
			}

			/* Check the size of the message is 20 (header) + 12 + 20 + 16 * 3 + 101 + 3 (padding) = 204 */
			CHECK( 0, msg_update_length ( msg ) );
			CHECK( 204, msgdata->msg_length );
			
			/* Add an octetstring from an enumerated constant */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumos" );
			{
				dict_object_t * type_model = NULL;
				dict_object_t * value_model = NULL;
				dict_type_enum_request_t request;
				
				CHECK( 0, msg_model ( avpi, &avp_model ) );
				CHECK( 0, dict_search ( DICT_TYPE, TYPE_OF_AVP, avp_model, &type_model ) );
				memset(&request, 0, sizeof(request));
				request.type_obj = type_model;
				request.search.enum_name = "os const test (waaad)";
				CHECK( 0, dict_search ( DICT_TYPE_ENUM, ENUM_BY_STRUCT, &request, &value_model ) );
				CHECK( 0, dict_getval ( value_model, &request.search ) );
				CHECK( 0, msg_avp_setvalue ( avpi, &request.search.enum_value ) );
				#if 0
				log_debug("AVP Enumuerated OctetString:\n");
				msg_dump_one(avpi);
				#endif
				/* Check the size is correct ( 12 for header + 5 for value ) */
				CHECK( 0, msg_update_length ( avpi ) );
				CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
				CHECK( 17, avpdata->avp_len );
			}
				
			/* Add an octetstring from an enumerated constant */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - enumos" );
			{
				dict_object_t * type_model = NULL;
				dict_object_t * value_model = NULL;
				dict_type_enum_request_t request;
				
				CHECK( 0, msg_model ( avpi, &avp_model ) );
				CHECK( 0, dict_search ( DICT_TYPE, TYPE_OF_AVP, avp_model, &type_model ) );
				memset(&request, 0, sizeof(request));
				request.type_obj = type_model;
				request.search.enum_name = "os const test (waa)";
				CHECK( 0, dict_search ( DICT_TYPE_ENUM, ENUM_BY_STRUCT, &request, &value_model ) );
				CHECK( 0, dict_getval ( value_model, &request.search ) );
				CHECK( 0, msg_avp_setvalue ( avpi, &request.search.enum_value ) );
				#if 0
				log_debug("AVP Enumuerated OctetString:\n");
				msg_dump_one(avpi);
				#endif
				/* Check the size is correct ( 12 for header + 3 for value ) */
				CHECK( 0, msg_update_length ( avpi ) );
				CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
				CHECK( 15, avpdata->avp_len );
			}
				

			/* Check the size of the message is 20 (header) + 12 + 20 + 16 * 3 + (101 + 3) + (17 + 3) + (15 + 1) = 240 */
			CHECK( 0, msg_update_length ( msg ) );
			CHECK( 240, msgdata->msg_length );
			
			/* Now test the grouped AVPs */	
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - grouped" );
			  ADD_AVP( avpi, MSG_BRW_LAST_CHILD, avpch, 73565, "AVP Test - os" );
			  {
				value.os.data = (unsigned char *)"12345678";
				value.os.len = 8;
				CHECK( 0, msg_avp_setvalue ( avpch, &value ) );
				#if 0
				log_debug("AVP octet string, '1234678'\n");
				msg_dump_one(avpch);
				#endif
				CHECK( 0, msg_update_length ( avpch ) );
				CHECK( 0, msg_avp_data ( avpch, &avpdata ) );
				CHECK( 20, avpdata->avp_len );
			  }
			  ADD_AVP( avpi, MSG_BRW_LAST_CHILD, avpch, 73565, "AVP Test - os" );
			  {
				value.os.data = (unsigned char *)"123456789";
				value.os.len = 9;
				CHECK( 0, msg_avp_setvalue ( avpch, &value ) );
				#if 0
				log_debug("AVP octet string, '12346789'\n");
				msg_dump_one(avpch);
				#endif
			  }
			
			/* Check the size is updated recursively: (gavp hdr: 12) + (avp1: 20) + (avp2: 21 + 3) = 56 */
			CHECK( 0, msg_update_length ( avpi ) );
			CHECK( 0, msg_avp_data ( avpi, &avpdata ) );
			CHECK( 56, avpdata->avp_len );
			
			/* Add another similar grouped AVP, to have lot of padding */
			ADD_AVP( msg, MSG_BRW_LAST_CHILD, avpi, 73565, "AVP Test - grouped" );
			  ADD_AVP( avpi, MSG_BRW_LAST_CHILD, avpch, 73565, "AVP Test - os" );
			  {
				value.os.data = (unsigned char *)"1";
				value.os.len = 1;
				CHECK( 0, msg_avp_setvalue ( avpch, &value ) );
			  }
			  ADD_AVP( avpi, MSG_BRW_LAST_CHILD, avpch, 73565, "AVP Test - os" );
			  {
				value.os.data = (unsigned char *)"1234567";
				value.os.len = 7;
				CHECK( 0, msg_avp_setvalue ( avpch, &value ) );
			  }
			
			/* Now check the global size of the message, if padding is correctly handled */
			/* size = 20 (header) + 12 + 20 + 16 * 3 + (101 + 3) + (17 + 3) + (15 + 1) 
			 *        + ( 12 + ( 20 + 21) + 3 )         # padding for the grouped AVP = 3
			 *        + ( 12 + ( (13 + 3) + 19 ) + 1 )  # and 1 for this one
			 * size = 240 + 56 + 48 = 344
			 */
			CHECK( 0, msg_update_length ( msg ) );
			#if 0
			msg_dump_walk(msg);
			#endif
			CHECK( 344, msgdata->msg_length );
			
			/* Set the application to the test application: 73566 */
			msgdata->msg_appl = 73566;
			
			/* Set the hop-by-hop ID to a random value: 0x4b44b41d */
			msgdata->msg_hbhid = 0x4b44b41d;
			/* Set the end-to-end ID to a random value: 0xe2ee2e1d */
			msgdata->msg_eteid = 0xe2ee2e1d;
		}
		
		/* Test the msg_bufferize function */
		{
			
			CHECK( 0, msg_bufferize( msg, &buf, NULL ) );
			
			/* Test the first bytes */
			CHECK( 0x01, buf[0] ); /* Version */
			CHECK( 0x00, buf[1] ); /* Length: 344 = 0x000158 */
			CHECK( 0x01, buf[2] );
			CHECK( 0x58, buf[3] );
			CHECK( 0x80, buf[4] ); /* flags: only "R" is set. */
			CHECK( 0x01, buf[5] ); /* Command code: 73573 = 0x011F65 */
			CHECK( 0x1F, buf[6] );
			CHECK( 0x65, buf[7] );
			CHECK( 0x00, buf[8] ); /* App ID: 73566 = 0x00011F5E */
			CHECK( 0x01, buf[9] ); 
			CHECK( 0x1F, buf[10] );
			CHECK( 0x5E, buf[11] );
			CHECK( 0x4b, buf[12] ); /* hop-by-hop id: 0x4b44b41d */
			CHECK( 0x44, buf[13] );
			CHECK( 0xb4, buf[14] );
			CHECK( 0x1d, buf[15] );
			CHECK( 0xe2, buf[16] ); /* end-to-end id: 0xe2ee2e1d */
			CHECK( 0xee, buf[17] );
			CHECK( 0x2e, buf[18] );
			CHECK( 0x1d, buf[19] );
			
			CHECK( 0x00, buf[20] ); /* First AVP (AVP Test - no vendor - f32) begin: code 73567 = 0x00011F5F */
			CHECK( 0x01, buf[21] );
			CHECK( 0x1F, buf[22] );
			CHECK( 0x5F, buf[23] );
			CHECK( 0x00, buf[24] ); /* flags: 0 */
			CHECK( 0x00, buf[25] ); /* length: 12 = 0x00000c */
			CHECK( 0x00, buf[26] );
			CHECK( 0x0C, buf[27] );
			CHECK( 0x40, buf[28] ); /* Value: 3.1415:  sign = '+' => most significant bit = 0 */
			CHECK( 0x49, buf[29] ); /* 2 <= 3.1415 < 4 => exponent = 2 => biaised (on 8 bits) = (decimal) 129 = (binary) 100 0000 0 */
			CHECK( 0x0e, buf[30] ); /* significand = (decimal) 0.57075 = (binary) 0.100 1001 0000 1110 0101 0110 */
			CHECK( 0x56, buf[31] ); /* total => 0100 0000 0100 1001 0000 1110 0101 0110 = (hexa) 40 49 0e 56*/
			
			/* The other AVPs will be tested by successful parsing... */
		}
		
		/* Now free the message, we keep only the buffer. */
		CHECK( 0, msg_free( msg, 1 ) );
		
	}
	
	/* Test the parsing of buffers and messages */
	{
		unsigned char * buf_cpy = NULL;
		msg_t * msg;
		
		#define CPYBUF() {			\
			buf_cpy = malloc(344);		\
			CHECK( buf_cpy ? 1 : 0, 1);	\
			memcpy(buf_cpy, buf, 344);	\
		}
		
		/* Test the msg_parse_buffer function */
		{
			CPYBUF();
			CHECK( EBADMSG, msg_parse_buffer( &buf_cpy, 340, &msg) );
			
			CPYBUF();
			CHECK( 0, msg_parse_buffer( &buf_cpy, 344, &msg) );
			#if 0
			msg_dump_walk(msg);
			#endif
			
			/* reinit the msg */
			CHECK( 0, msg_free ( msg, 1 ) );
				
		}
		
		/* Test the msg_parse_dict function */
		{
			/* Test with an unknown command code */
			{
				CPYBUF();
				
				/* Change the command-code */
				buf_cpy[5] = 0x11;
				CHECK( 0, msg_parse_buffer( &buf_cpy, 344, &msg) );
				CHECK( ENOTSUP, msg_parse_dict( msg ) );
				
				/* reset */
				CHECK( 0, msg_free ( msg, 1 ) );
			}
			
			/* Test with an unknown Mandatory AVP */
			{
				CPYBUF();
				
				buf_cpy[20] = 0x11;	/* New AVP code = 0x11011F5F, undefined */
				buf_cpy[24] = 0x40; 	/* Add the 'M' flag */
				
				/* Check that we cannot support this message now */
				CHECK( 0, msg_parse_buffer( &buf_cpy, 344, &msg) );
				CHECK( ENOTSUP, msg_parse_dict( msg ) );
				
				/* reset */
				CHECK( 0, msg_free ( msg, 1 ) );
			}
			
			/* Test with an unknown optional AVP */
			{
				CPYBUF();
				
				buf_cpy[20] = 0x11;	/* New AVP code = 0x11011F5F, undefined */
				
				/* Check that we can support this message now */
				CHECK( 0, msg_parse_buffer( &buf_cpy, 344, &msg) );
				CHECK( 0, msg_parse_dict( msg ) );
				
				#if 0
				msg_dump_walk(msg);
				#endif
				
				/* Now test the "getraw" function */
				{
					msg_avp_t * avp;
					unsigned char * raw;
					
					CHECK( 0, msg_browse( msg, MSG_BRW_FIRST_CHILD, &avp, NULL) );
					
					CHECK( 0, msg_avp_getrawdata ( avp, &raw ) );
					
					CHECK( 0, memcmp(raw, buf + 28, 4) ); /* This is the representation of "3.1415" */
					
				}
				
				/* reset */
				CHECK( 0, msg_free ( msg, 1 ) );
			}
			
			CHECK( 0, msg_parse_buffer( &buf, 344, &msg) );
			CHECK( 0, msg_parse_dict( msg ) );
			#if 0
			msg_dump_walk(msg);
			#endif
		}
		
		/* Now test the msg_parse_rule function */
		{
			dict_object_t * rule;
			
			CHECK( 0, msg_parse_rules( msg, &rule ) );
			
			/* Use the "AVP Test - rules" AVP to test the rules */
			{
				msg_avp_t * tavp = NULL;
				msg_avp_t * tempavp = NULL;
				msg_avp_t * childavp = NULL;
				
				ADD_AVP( msg, MSG_BRW_LAST_CHILD, tavp, 73565, "AVP Test - rules" );
				
				/* Create a conforming message first */
				ADD_AVP( tavp, MSG_BRW_LAST_CHILD, childavp,     0, "AVP Test - no vendor - f32" );
				ADD_AVP( tavp, MSG_BRW_LAST_CHILD, childavp, 73565, "AVP Test - i64" );
				ADD_AVP( tavp, MSG_BRW_LAST_CHILD, childavp, 73565, "AVP Test - enumi32" );
				ADD_AVP( tavp, MSG_BRW_LAST_CHILD, childavp, 73565, "AVP Test - os" );
				ADD_AVP( tavp, MSG_BRW_LAST_CHILD, childavp, 73565, "AVP Test - os" );
				ADD_AVP( tavp, MSG_BRW_LAST_CHILD, childavp, 73565, "AVP Test - grouped" );
				
				/* Check the message is still conform */
				CHECK( 0, msg_parse_rules( msg, &rule ) );
				
				/* The first avp is optional in fixed position, so remove it and check the message is still OK */
				CHECK( 0, msg_browse ( tavp, MSG_BRW_FIRST_CHILD, &childavp, NULL) );
				CHECK( 0, msg_free ( childavp, 0 ) );
				CHECK( 0, msg_parse_rules( msg, &rule ) );
				ADD_AVP( tavp, MSG_BRW_FIRST_CHILD, childavp,     0, "AVP Test - no vendor - f32" );
				
				
				/* Now break some rules and check it is detected */
				#define CHECK_CONFLICT( _msg, _ruleavp, _avpvendor )		{				\
					dict_object_t * _rule;									\
					CHECK( EBADMSG,  msg_parse_rules( msg, &_rule ) );					\
					if ((_ruleavp) == NULL) {								\
						CHECK( NULL, _rule);								\
					} else {										\
						dict_rule_data_t _ruledata;							\
						dict_object_t * _avp;								\
						dict_avp_request_t _req = { (_avpvendor), 0, (_ruleavp) };			\
						CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME_AND_VENDOR, &_req, &_avp ) );	\
						CHECK( 0, dict_getval( _rule, &_ruledata ) );					\
						CHECK( _avp, _ruledata.rule_avp );						\
					}											\
				}

				{
					/* Test the FIXED_HEAD rules positions: add another AVP before the third */
					CHECK( 0, msg_browse ( tavp, MSG_BRW_FIRST_CHILD, &tempavp, NULL) ); /* tempavp is the novendor avp */
					CHECK( 0, msg_browse ( tempavp, MSG_BRW_NEXT, &tempavp, NULL) );     /* tempavp is the i64 avp */
					ADD_AVP( tempavp, MSG_BRW_NEXT, childavp, 73565, "AVP Test - os" );
					
					CHECK_CONFLICT( msg, "AVP Test - enumi32", 73565 );
					
					/* Now remove this AVP */
					CHECK( 0, msg_free ( childavp, 0 ) );
				}
				{
					/* Remove the third AVP, same rule must conflict */
					CHECK( 0, msg_browse ( tempavp, MSG_BRW_NEXT, &childavp, NULL) );     /* childavp is the enumi32 avp */
					CHECK( 0, msg_free ( childavp, 0 ) );
					
					CHECK_CONFLICT( msg, "AVP Test - enumi32", 73565 );
					
					/* Add the AVP back */
					ADD_AVP( tempavp, MSG_BRW_NEXT, childavp, 73565, "AVP Test - enumi32" );
				}
				
				{
					/* Test the minimum value in the REQUIRED rule: delete one of the os AVPs */
					CHECK( 0, msg_browse ( childavp, MSG_BRW_NEXT, &tempavp, NULL) );     /* tempavp is the os avp */
					CHECK( 0, msg_free ( tempavp, 0 ) );
					
					CHECK_CONFLICT( msg, "AVP Test - os", 73565 );	/* The rule requires at least 2 AVP, we have only 1 */
					
					/* Now add this AVP */
					ADD_AVP( childavp, MSG_BRW_NEXT, tempavp, 73565, "AVP Test - os" );
				}
				{
					/* Test the maximum value in the REQUIRED rule: add more of the os AVPs */
					ADD_AVP( childavp, MSG_BRW_NEXT, tempavp, 73565, "AVP Test - os" );
					ADD_AVP( childavp, MSG_BRW_NEXT, tempavp, 73565, "AVP Test - os" );
					
					CHECK_CONFLICT( msg, "AVP Test - os", 73565 );	/* The rule requires at most 3 AVP, we have 4 */
					
					/* Now delete these AVP */
					CHECK( 0, msg_free ( tempavp, 0 ) );
					CHECK( 0, msg_browse ( childavp, MSG_BRW_NEXT, &tempavp, NULL) );
					CHECK( 0, msg_free ( tempavp, 0 ) );
				}
				
				{
					/* Test the maximum value in the OPTIONAL rule: add 2 enumos AVPs */
					ADD_AVP( childavp, MSG_BRW_NEXT, tempavp, 73565, "AVP Test - enumos" );
					
					/* The message is still conform */
					CHECK( 0, msg_parse_rules( msg, &rule ) );
					
					/* Now break the rule */
					ADD_AVP( childavp, MSG_BRW_NEXT, tempavp, 73565, "AVP Test - enumos" );
					
					CHECK_CONFLICT( msg, "AVP Test - enumos", 73565 );	/* The rule requires at most 3 AVP, we have 4 */
					
					/* Now delete this AVP */
					CHECK( 0, msg_free ( tempavp, 0 ) );
				}
				
				{
					/* Test the RULE_FIXED_TAIL rules positions: add another AVP at the end */
					ADD_AVP( tavp, MSG_BRW_LAST_CHILD, childavp, 73565, "AVP Test - os" );
					
					CHECK_CONFLICT( msg, "AVP Test - grouped", 73565 );
					
					/* Now remove this AVP */
					CHECK( 0, msg_free ( childavp, 0 ) );
				}
			}
		}
	}
	
	/* Test the msg_avp_value_interpret and msg_avp_value_encode functions. use the Address type and Host-IP-Address AVPs */
	{
		dict_object_t * cer_model = NULL;
		msg_t * cer = NULL;
		
		dict_object_t * hia_model = NULL;
		msg_avp_t *avp4, *avp6;
		#define TEST_IP4 "192.168.100.101"
		char buf4[INET_ADDRSTRLEN];
		#define	TEST_IP6 "1111:2222:3333:4444:1234:5678:9abc:def0"
		char buf6[INET6_ADDRSTRLEN];
		
		struct sockaddr_storage ss;
		struct sockaddr_in  sin,  *psin;
		struct sockaddr_in6 sin6, *psin6;
		
		/* Find the CER dictionary object */
		CHECK( 0, dict_search ( DICT_COMMAND, CMD_BY_NAME, "Capabilities-Exchange-Request", &cer_model ) );

		/* Now find the Host-IP-Address dictionary object */
		CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME, "Host-IP-Address", &hia_model ) );

		/* Create the msg instance */
		CHECK( 0, msg_new ( cer_model, 0, &cer ) );

		/* Create the avp instances */
		CHECK( 0, msg_avp_new ( hia_model, 0, &avp4 ) );
		CHECK( 0, msg_avp_new ( hia_model, 0, &avp6 ) );
		
		/* Set the value of the IP avp */
		sin.sin_family = AF_INET;
		CHECK( 1, inet_pton( AF_INET, TEST_IP4, &sin.sin_addr.s_addr ) );
		CHECK( 0, msg_avp_value_encode ( &sin, avp4 ) );
		
		/* Set the value of the IP6 avp */
		sin6.sin6_family = AF_INET6;
		CHECK( 1, inet_pton( AF_INET6, TEST_IP6, &sin6.sin6_addr.s6_addr ) );
		CHECK( 0, msg_avp_value_encode ( &sin6, avp6 ) );
		
		/* Add these AVPs in the message */
		CHECK( 0, msg_avp_add( cer, MSG_BRW_LAST_CHILD, avp4) );
		CHECK( 0, msg_avp_add( cer, MSG_BRW_LAST_CHILD, avp6) );
		
		/* Create the buffer for this message */
		CHECK( 0, msg_bufferize( cer, &buf, NULL ) );
		
		/* Now free the message, we keep only the buffer. */
		CHECK( 0, msg_free( cer, 1 ) );
		
		/* Check the content of the buffer is correct (skip command header) */
		CHECK( 0x00, buf[20] ); /* First AVP (IP4) begins: code 257 = 0x00000101 */
		CHECK( 0x00, buf[21] );
		CHECK( 0x01, buf[22] );
		CHECK( 0x01, buf[23] );
		CHECK( 0x40, buf[24] ); /* flags: M */
		CHECK( 0x00, buf[25] ); /* length: 8+6 = 0x00000e */
		CHECK( 0x00, buf[26] );
		CHECK( 0x0E, buf[27] );
		CHECK( 0x00, buf[28] ); /* Value: AddressType 1 */
		CHECK( 0x01, buf[29] ); 
		CHECK(  192, buf[30] ); /* 192.168.100.101 */
		CHECK(  168, buf[31] ); 
		CHECK(  100, buf[32] ); 
		CHECK(  101, buf[33] );
		
		CHECK( 0x00, buf[34] ); /* Padding */
		CHECK( 0x00, buf[35] );
		
		CHECK( 0x00, buf[36] ); /* Second AVP (IP6) begins: code 257 = 0x00000101 */
		CHECK( 0x00, buf[37] );
		CHECK( 0x01, buf[38] );
		CHECK( 0x01, buf[39] );
		CHECK( 0x40, buf[40] ); /* flags: M */
		CHECK( 0x00, buf[41] ); /* length: 8+18 = 0x00001a */
		CHECK( 0x00, buf[42] );
		CHECK( 0x1A, buf[43] );
		CHECK( 0x00, buf[44] ); /* Value: AddressType 2 */
		CHECK( 0x02, buf[45] ); 
		CHECK( 0x11, buf[46] ); /* 1111:2222:3333:4444:1234:5678:9abc:def0 */
		CHECK( 0x11, buf[47] ); 
		CHECK( 0x22, buf[48] ); 
		CHECK( 0x22, buf[49] );
		CHECK( 0x33, buf[50] );
		CHECK( 0x33, buf[51] );
		CHECK( 0x44, buf[52] );
		CHECK( 0x44, buf[53] );
		CHECK( 0x12, buf[54] );
		CHECK( 0x34, buf[55] );
		CHECK( 0x56, buf[56] );
		CHECK( 0x78, buf[57] );
		CHECK( 0x9a, buf[58] );
		CHECK( 0xbc, buf[59] );
		CHECK( 0xde, buf[60] );
		CHECK( 0xf0, buf[61] );
		
		/* Ok, now let's recreate the message */
		CHECK( 0, msg_parse_buffer( &buf, 64, &cer) );
		CHECK( 0, msg_parse_dict( cer ) );
		
		/* Get the pointers to the first and last AVP */
		CHECK( 0, msg_browse( cer, MSG_BRW_FIRST_CHILD, &avp4, NULL) );
		CHECK( 0, msg_browse( cer, MSG_BRW_LAST_CHILD,  &avp6, NULL) );
		
		/* Try and interpret the data in the AVPs */
		CHECK( 0, msg_avp_value_interpret ( avp4, &ss ) );
		psin = (struct sockaddr_in *)&ss;
		CHECK( AF_INET, psin->sin_family );
		CHECK( 0, (inet_ntop( AF_INET, &psin->sin_addr.s_addr, buf4, sizeof(buf4) ) == NULL) ? errno : 0 );
		CHECK( 0, strcmp( buf4, TEST_IP4 ) );
		
		CHECK( 0, msg_avp_value_interpret ( avp6, &ss ) );
		psin6 = (struct sockaddr_in6 *)&ss;
		CHECK( AF_INET6, psin6->sin6_family );
		CHECK( 0, (inet_ntop( AF_INET6, &psin6->sin6_addr.s6_addr, buf6, sizeof(buf6) ) == NULL) ? errno : 0 );
		CHECK( 0, strcasecmp( buf6, TEST_IP6 ) );
		
		/* Ok, it's done */
		CHECK( 0, msg_free( cer, 1 ) );
	}

	/* That's all for the tests yet */
	PASSTEST();
} 
	
